/** \file Refinement_DOFView.h */

#ifndef REFINEMENT_DOFVIEW_H
#define REFINEMENT_DOFVIEW_H

#include "Refinement.h"
#include "Views.h"
#include "MeshFunction_Level.h"

using namespace AMDiS;


/** \brief
 * Refinement structure to perform local anisotropic refinement depending
 * on a refineFunction that defines the local refinement level to a given
 * value, like phase-value or distance-value.
 **/
class RefinementLevelDOFView : public RefinementLevel<double, int>
{
public:  
    typedef double T;
    
public:
  RefinementLevelDOFView(const FiniteElemSpace *feSpace_, 
			 MeshRefinementFunction<T, int>* refineFct_, 
			 DOFView<double> *vec_)
  : RefinementLevel<T, int>(feSpace_, refineFct_), 
    vec(vec_) { };

  void markElements(Flag &markFlag) 
  {
    FUNCNAME("RefinementLevelDOFView::markElements()");

    T meanValue; nullify(meanValue);
    WorldVector<double> centroid;

    bool elMarkRefine = false, elMarkCoarsen = false;
    
    int dim = mesh->getDim();

    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, traverseFlag);
    MacroElement* foundMacro = new MacroElement(dim);
    bool found = false;
    while (elInfo) {
      centroid = elInfo->getCoord(0);
      for (int i = 1; i < dim+1; i++)
	centroid += elInfo->getCoord(i);
      centroid *= 1.0/(dim+1.0);
      
      meanValue = evalAtPoint<T>(*vec, centroid);

      int refineLevel = (*refineFct)(meanValue);
      int oldLevel = elInfo->getLevel();
      elInfo->getElement()->setMark( calcMark(refineLevel, oldLevel) );

      elMarkRefine |= elInfo->getElement()->getMark() >= 1;
      elMarkCoarsen |= elInfo->getElement()->getMark() <= -1;
      elInfo = stack.traverseNext(elInfo);
    }
		
    markFlag = 0;
    if (elMarkRefine) 
      markFlag = 1;
    if (elMarkCoarsen) 
      markFlag |= 2;
  };
  
private:

  DOFView<double>* vec;
};



/** \brief
 * Specialization of RefinementLevel, that uses an arbitrary container 'data'
 * that gives the phase-field value or the distance value, or something that 
 * can be handled by the reine-function.
 * For the Container a free function evalAtPoint must be implemented, 
 * e.g. DOFVector, DOFView, AbstractFunction, POD-Type.
 **/
template<typename Container>
class RefinementLevel1 : public RefinementLevel<double, int>
{
public:  
    typedef double T;
    
public:
  RefinementLevel1(const FiniteElemSpace *feSpace_, 
		   MeshRefinementFunction<T, int>* refineFct_, 
		   Container &data_)
  : RefinementLevel<T, int>(feSpace_, refineFct_), 
    data(data_) { };

  void markElements(Flag &markFlag) 
  {
    FUNCNAME("RefinementLevelDOFView::markElements()");

    bool elMarkRefine = false, elMarkCoarsen = false;
    
    int dim = mesh->getDim();

    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, traverseFlag);

    while (elInfo) {
      WorldVector<double> centroid = elInfo->getCoord(0);
      for (int i = 1; i < dim+1; i++)
	centroid += elInfo->getCoord(i);
      centroid *= 1.0/(dim+1.0);
      
      typename ValueType<Container>::type meanValue = evalAtPoint<typename ValueType<Container>::type>(data, centroid, elInfo);

      int refineLevel = (*refineFct)(meanValue);
      int oldLevel = elInfo->getLevel();
      elInfo->getElement()->setMark( calcMark(refineLevel, oldLevel) );

      elMarkRefine |= elInfo->getElement()->getMark() >= 1;
      elMarkCoarsen |= elInfo->getElement()->getMark() <= -1;
      elInfo = stack.traverseNext(elInfo);
    }
		
    markFlag = 0;
    if (elMarkRefine) 
      markFlag = 1;
    if (elMarkCoarsen) 
      markFlag |= 2;
  };
  
private:
  Container& data;
};


/** \brief
 * Specialization of RefinementLevel, that uses two arbitrary containers 'data1', 'data2'
 * that give the phase-field values or the distance values, or something that 
 * can be handled by the reine-function.
 * For the Containers free functions evalAtPoint must be implemented, 
 * e.g. DOFVector, DOFView, AbstractFunction, POD-Type.
 **/
template<typename Container1, typename Container2>
class RefinementLevel2 : public RefinementLevel<boost::tuple<double, double>, int>
{
public:  
    typedef boost::tuple<double, double> T;
    
public:
  RefinementLevel2(const FiniteElemSpace *feSpace_, 
		   MeshRefinementFunction<T, int>* refineFct_, 
		   Container1 &data1_,
		   Container2 &data2_)
  : RefinementLevel<T, int>(feSpace_, refineFct_), 
    data1(data1_),
    data2(data2_) { };

  void markElements(Flag &markFlag) 
  {
    FUNCNAME("RefinementLevelDOFView::markElements()");

    bool elMarkRefine = false, elMarkCoarsen = false;
    
    int dim = mesh->getDim();

    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, traverseFlag);

    while (elInfo) {
      WorldVector<double> centroid = elInfo->getCoord(0);
      for (int i = 1; i < dim+1; i++)
	centroid += elInfo->getCoord(i);
      centroid *= 1.0/(dim+1.0);
      
      typename ValueType<Container1>::type meanValue1 = evalAtPoint<typename ValueType<Container1>::type>(data1, centroid, elInfo);
      typename ValueType<Container1>::type meanValue2 = evalAtPoint<typename ValueType<Container2>::type>(data2, centroid, elInfo);

      int refineLevel = (*refineFct)(boost::tuples::make_tuple(meanValue1, meanValue2));
      int oldLevel = elInfo->getLevel();
      elInfo->getElement()->setMark( calcMark(refineLevel, oldLevel) );

      elMarkRefine |= elInfo->getElement()->getMark() >= 1;
      elMarkCoarsen |= elInfo->getElement()->getMark() <= -1;
      elInfo = stack.traverseNext(elInfo);
    }
		
    markFlag = 0;
    if (elMarkRefine) 
      markFlag |= 1;
    if (elMarkCoarsen) 
      markFlag |= 2;
  };
  
private:
  Container1& data1;
  Container2& data2;
};


/** \brief
 * Specialization of RefinementLevel, that uses two arbitrary containers 'data1', 'data2'
 * that give the phase-field values or the distance values, or something that 
 * can be handled by the reine-function.
 * For the Containers free functions evalAtPoint must be implemented, 
 * e.g. DOFVector, DOFView, AbstractFunction, POD-Type.
 **/
template<typename Container1, typename Container2, typename Container3>
class RefinementLevel3 : public RefinementLevel<boost::tuple<double, double, double>, int>
{
public:  
    typedef boost::tuple<double, double, double> T;
    
public:
  RefinementLevel3(const FiniteElemSpace *feSpace_, 
		   MeshRefinementFunction<T, int>* refineFct_, 
		   Container1 &data1_,
		   Container2 &data2_,
		   Container3 &data3_)
  : RefinementLevel<T, int>(feSpace_, refineFct_), 
    data1(data1_),
    data2(data2_),
    data3(data3_) { };

  void markElements(Flag &markFlag)
  {
    FUNCNAME("RefinementLevelDOFView::markElements()");

    bool elMarkRefine = false, elMarkCoarsen = false;
    
    int dim = mesh->getDim();

    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, traverseFlag);

    while (elInfo) {
      WorldVector<double> centroid = elInfo->getCoord(0);
      for (int i = 1; i < dim+1; i++)
	centroid += elInfo->getCoord(i);
      centroid *= 1.0/(dim+1.0);
      
      typename ValueType<Container1>::type meanValue1 = evalAtPoint<typename ValueType<Container1>::type>(data1, centroid, elInfo);
      typename ValueType<Container2>::type meanValue2 = evalAtPoint<typename ValueType<Container2>::type>(data2, centroid, elInfo);
      typename ValueType<Container3>::type meanValue3 = evalAtPoint<typename ValueType<Container3>::type>(data3, centroid, elInfo);

      int refineLevel = (*refineFct)(boost::tuples::make_tuple(meanValue1, meanValue2, meanValue3));
      int oldLevel = elInfo->getLevel();
      elInfo->getElement()->setMark( calcMark(refineLevel, oldLevel) );

      elMarkRefine |= elInfo->getElement()->getMark() >= 1;
      elMarkCoarsen |= elInfo->getElement()->getMark() <= -1;
      elInfo = stack.traverseNext(elInfo);
    }
		
    markFlag = 0;
    if (elMarkRefine) 
      markFlag = 1;
    if (elMarkCoarsen) 
      markFlag |= 2;
  };
  
private:
  Container1& data1;
  Container2& data2;
  Container3& data3;
};


/** \brief
 * Refinement structure to perform local anisotropic refinement depending
 * on a refineFunction that defines the local refinement level to a given
 * list of values, like phase-values or distance-values.
 **/
class RefinementLevelDOF2 : public RefinementLevel< boost::tuple<double, double>, int >
{
public:  
    typedef boost::tuple<double, double> value_Type;
    
public:
  RefinementLevelDOF2(
      const FiniteElemSpace *feSpace_, 
      MeshRefinementFunction< value_Type, int >* refineFct_, 
      DOFVector<double>* vec0_, 
      DOFVector<double>* vec1_)
  : RefinementLevel< value_Type, int >(feSpace_, refineFct_),
    vec0(vec0_),
    vec1(vec1_) { };

  void markElements(Flag &markFlag) 
  {
    FUNCNAME("RefinementLevelDOFViewAndDOFVector::markElements()");

    const DOFAdmin* admin = feSpace->getAdmin();
    const BasisFunction *basFcts = feSpace->getBasisFcts();
    int numBasFcts = basFcts->getNumber();

    DegreeOfFreedom *localIndices = new DegreeOfFreedom[numBasFcts];

    bool elMarkRefine = false, elMarkCoarsen = false;

    ElementFunctionDOFVec<double> elFct0(vec0);
    ElementFunctionDOFVec<double> elFct1(vec1);
    DimVec<double> lambda(mesh->getDim(), NO_INIT);
    for (int i = 0; i < lambda.getSize(); ++i)
      lambda[i] = 1.0 / lambda.getSize();

    int dim = mesh->getDim();
    Flag traverseFlag = Mesh::CALL_LEAF_EL;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, traverseFlag);
    while(elInfo) {
      // eval DOFVector0
      elFct0.setElInfo(elInfo);
      double meanValue0 = elFct0(lambda);
      
      // eval DOFVector1
      elFct1.setElInfo(elInfo);
      double meanValue1 = elFct1(lambda);
      
      int refineLevel = (*refineFct)(boost::tuples::make_tuple(meanValue0, meanValue1));
      int oldLevel = elInfo->getLevel();
      elInfo->getElement()->setMark( calcMark(refineLevel, oldLevel) );

      elMarkRefine |= elInfo->getElement()->getMark() == 1;
      elMarkCoarsen |= elInfo->getElement()->getMark() == -1;
      elInfo = stack.traverseNext(elInfo);
    }
      
    markFlag= 0;
    if(elMarkRefine) 
      markFlag = 1;
    if(elMarkCoarsen) 
      markFlag |= 2;
    
    delete [] localIndices;
  };
  
  
private:
  
  DOFVector<double>* vec0;
  DOFVector<double>* vec1;
};

/** \brief
 * Refinement structure to perform local anisotropic refinement depending
 * on a refineFunction that defines the local refinement level to a given
 * list of values, like phase-values or distance-values.
 **/
class RefinementLevelDOFViewAndDOFVector : public RefinementLevel< boost::tuple<double, double>, int >
{
public:  
    typedef boost::tuple<double, double> value_Type;
    
public:
  RefinementLevelDOFViewAndDOFVector(
      const FiniteElemSpace *feSpace_, 
      MeshRefinementFunction< value_Type, int >* refineFct_, 
      DOFView<double>* vec0_, 
      DOFVector<double>* vec1_)
  : RefinementLevel< value_Type, int >(feSpace_, refineFct_),
    vec0(vec0_),
    vec1(vec1_) { };

  void markElements(Flag &markFlag) 
  {
    FUNCNAME("RefinementLevelDOFViewAndDOFVector::markElements()");

    const DOFAdmin* admin = feSpace->getAdmin();
    const BasisFunction *basFcts = feSpace->getBasisFcts();
    int numBasFcts = basFcts->getNumber();

    DegreeOfFreedom *localIndices = new DegreeOfFreedom[numBasFcts];

    bool elMarkRefine = false, elMarkCoarsen = false;

    ElementFunctionDOFVec<double> elFct(vec1);
    DimVec<double> lambda(mesh->getDim(), NO_INIT);
    for (int i = 0; i < lambda.getSize(); ++i)
      lambda[i] = 1.0 / lambda.getSize();

    int dim = mesh->getDim();
    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, traverseFlag);
    WorldVector<double> border; border.set(0.0); border[0] = 0.5;
    while (elInfo) {
      WorldVector<double> centroid = elInfo->getCoord(0);
      for (int i = 1; i < dim+1; i++)
	centroid += elInfo->getCoord(i);
      centroid *= 1.0/(dim+1.0);
      
      // eval DOFView
      double meanValue0 = evalAtPoint<double>(*vec0, centroid);   
      // eval DOFVector
      elFct.setElInfo(elInfo);
      double meanValue1 = elFct(lambda);
      
      int refineLevel = (*refineFct)(boost::tuples::make_tuple(meanValue0, meanValue1));      
      int oldLevel = elInfo->getLevel();
      elInfo->getElement()->setMark( calcMark(refineLevel, oldLevel) );

      elMarkRefine |= elInfo->getElement()->getMark() == 1;
      elMarkCoarsen |= elInfo->getElement()->getMark() == -1;
      elInfo = stack.traverseNext(elInfo);
    }
      
    markFlag = 0;
    if(elMarkRefine) 
      markFlag = 1;
    if(elMarkCoarsen) 
      markFlag |= 2;
    
    delete [] localIndices;
  }
  
private:
  
  DOFView<double>* vec0;
  DOFVector<double>* vec1;
};



/** \brief
 * Implementation of PhaseFieldRefinementBase. Provides the method operator()
 * that defines the regions for interface refinement and domain refinement, etc.
 **/
class PhaseField2Refinement : public PhaseFieldRefinementBase<boost::tuple<double, double> >
{
public:

    PhaseField2Refinement(Mesh* mesh_) : PhaseFieldRefinementBase<boost::tuple<double, double> >(mesh_) {};
    
    int operator()(const boost::tuple<double, double>& phases) const {
      int result0= lOuter;
      if (minPhase < phases.get<0>() && phases.get<0>() < maxPhase)
        result0= lInterface;         // auf dem Interface
      else if (phases.get<0>() > minOuterPhase && phases.get<0>() <= minPhase)
        result0= std::max(lOuter, static_cast<int>(floor((lOuter+lInterface)/2.0)));
      else if (phases.get<0>() < maxOuterPhase && phases.get<0>() >= maxPhase)
        result0= std::max(lInner, static_cast<int>(floor((lInner+lInterface)/2.0)));
      else if (phases.get<0>() > (minPhase+maxPhase)/2.0)
        result0= lInner;
      
      int result1= lOuter;
      if (minPhase < phases.get<1>() && phases.get<1>() < maxPhase)
        result1= lInterface;         // auf dem Interface
      else if (phases.get<1>() > minOuterPhase && phases.get<1>() <= minPhase)
        result1= std::max(lOuter, static_cast<int>(floor((lOuter+lInterface)/2.0)));
      else if (phases.get<1>() < maxOuterPhase && phases.get<1>() >= maxPhase)
        result1= std::max(lInner, static_cast<int>(floor((lInner+lInterface)/2.0)));
      else if (phases.get<1>() > (minPhase+maxPhase)/2.0)
        result1= lInner;
      
      return std::max(result0, result1);
    }
};


/** \brief
 * Implementation of PhaseFieldRefinementBase. Provides the method operator()
 * that defines the regions for interface refinement and domain refinement, etc.
 **/
class PhaseField2Refinement_simple : public PhaseFieldRefinementBase<boost::tuple<double, double> >
{
public:

    PhaseField2Refinement_simple(Mesh* mesh_) : PhaseFieldRefinementBase<boost::tuple<double, double> >(mesh_) {};
    
    int operator()(const boost::tuple<double, double>& phases) const {
      int result0 = lOuter;
      if (minPhase < phases.get<0>() && phases.get<0>() < maxPhase)
        result0 = lInterface;         // auf dem Interface vom Phase 1
      
      int result1 = lOuter;
      if (minPhase < phases.get<1>() && phases.get<1>() < maxPhase)
        result1 = lInterface;         // auf dem Interface von Phase 2
      
      return std::max(result0, result1);
    }
};

#endif // REFINEMENT_DOFVIEW_H
