/** \file CahnHilliard_RB.h */

#ifndef CAHN_HILLIARD_RB_H
#define CAHN_HILLIARD_RB_H

#include "AMDiS.h"
#include "BaseProblem_RB.h"
#include "ExtendedProblemStat.h"
#include "HL_SignedDistTraverse.h"

using namespace AMDiS;

class CahnHilliard_RB : public BaseProblem_RB
{
public: // definition of types

  typedef BaseProblem_RB super;

public: // public methods

  CahnHilliard_RB(const std::string &name_);
  ~CahnHilliard_RB();

  virtual void initData();
  virtual void solveInitialProblem(AdaptInfo *adaptInfo);

  double getEpsilon() { return eps; }
  int getDoubleWellType() { return doubleWell; }

protected: // protected methods

  virtual void fillOperators();
  virtual void fillBoundaryConditions() {}

protected: // protected variables

  HL_SignedDistTraverse *reinit;

  bool useMobility;

  unsigned dim;

  int doubleWell;

  double gamma;
  double eps;
  double minusEps;
  double epsInv;
  double minusEpsInv;
  double epsSqr;
  double minusEpsSqr;
};


/** \brief
 * Abstract function for Cahn-Hilliard mobility
 */
class MobilityCH0 : public AbstractFunction<double,double>
{
  public:
    MobilityCH0(double gamma_=1.0) : 
      AbstractFunction<double,double>(4), 
      gamma(gamma_),
      delta(1.e-6) { }
    double operator()(const double &ch) const 
    {
      double phase = std::max(0.0, std::min(1.0, ch));
      double mobility = 0.25*sqr(phase)*sqr(phase-1.0);
      return gamma * std::max(mobility, delta);
    }

  protected:
    double gamma;
    double delta;
};

class MobilityCH1 : public AbstractFunction<double,double>
{
  public:
    MobilityCH1(double gamma_=1.0) :
      AbstractFunction<double,double>(4),
      gamma(gamma_),
      delta(1.e-6) { }
    double operator()(const double &ch) const
    {
      double phase = std::max(-1.0, std::min(1.0, ch));
      double mobility = 0.25*sqr(sqr(phase)-1.0);
      return gamma * std::max(mobility, delta);
    }

  protected:
    double gamma;
    double delta;
};

class ScaleFunctor : public AbstractFunction<double,double>
{
  public:
    ScaleFunctor(double factor_=1.0) :
      AbstractFunction<double,double>(1),
      factor(factor_) { }
    double operator()(const double &ch) const
    {
      return factor * ch;
    }

  protected:
    double factor;
};

class MobilityCH0Diff : public BinaryAbstractFunction<WorldVector<double>,double,WorldVector<double> >
{
  public:
    MobilityCH0Diff(double factor_=1.0) :
      BinaryAbstractFunction<WorldVector<double>,double,WorldVector<double> >(4),
      factor(factor_) { }
    WorldVector<double> operator()(const double &c, const WorldVector<double>& grdMu) const
    {
      double phase = std::max(0.0, std::min(1.0, c));
      return (factor * (sqr(phase)*phase - 1.5*sqr(phase) + 0.5*phase)) * grdMu;
    }

  protected:
    double factor;
};

class MobilityCH1Diff : public BinaryAbstractFunction<WorldVector<double>,double,WorldVector<double> >
{
  public:
    MobilityCH1Diff(double factor_=1.0) :
      BinaryAbstractFunction<WorldVector<double>,double,WorldVector<double> >(4),
      factor(factor_) { }
    WorldVector<double> operator()(const double &c, const WorldVector<double>& grdMu) const
    {
      double phase = std::max(-1.0, std::min(1.0, c));
      return (factor * (sqr(phase)*phase - phase)) * grdMu;
    }

  protected:
    double factor;
};

class DoubleWell0 : public AbstractFunction<double,double>
{
  public:
    DoubleWell0(double factor_=1.0) : 
      AbstractFunction<double,double>(4), 
      factor(factor_) { }
      
    double operator()(const double &c) const 
    {
      return factor * (sqr(c)*c - 1.5*sqr(c) + 0.5*c);
    }

  protected:
    double factor;
};

class DoubleWell1 : public AbstractFunction<double,double>
{
  public:
    DoubleWell1(double factor_=1.0) : 
      AbstractFunction<double,double>(4), 
      factor(factor_) { }
      
    double operator()(const double &c) const
    {
      return factor * (sqr(c)*c - c);
    }

  protected:
    double factor;
};

class DoubleWell0Diff : public AbstractFunction<double,double>
{
  public:
    DoubleWell0Diff(double factor_=1.0) : 
      AbstractFunction<double,double>(4), 
      factor(factor_) { }
      
    double operator()(const double &c) const 
    {
      return factor * (3.0*sqr(c) - 3.0*c + 0.5);
    }

  protected:
    double factor;
};

class DoubleWell1Diff : public AbstractFunction<double,double>
{
  public:
    DoubleWell1Diff(double factor_=1.0) : 
      AbstractFunction<double,double>(4), 
      factor(factor_) { }
      
    double operator()(const double &c) const
    {
      return factor * (3.0*sqr(c) - 1.0);
    }

  protected:
    double factor;
};

#endif // CAHN_HILLIARD_RB_H
