/** \file LinearElasticityPhase.h */

#ifndef LINEAR_ELASTICITY_H
#define LINEAR_ELASTICITY_H

#include "AMDiS.h"
#include "BaseProblem.h"
#include "ExtendedProblemStat.h"

using namespace AMDiS;

/** \ingroup Elasticity
 * \brief
 * Elastodynamics – the wave equation
 */
class LinearElasticityPhase : public BaseProblem<ExtendedProblemStat>
{
public: // typedefs
  
  typedef BaseProblem<ExtendedProblemStat> super;
  
public: // methods

  LinearElasticityPhase(const std::string &name_);

  /// if initialVelocityIsSet is true delete initialX and initialY
  ~LinearElasticityPhase();

  virtual void initData();
  virtual void solveInitialProblem(AdaptInfo *adaptInfo);

  virtual void transferInitialSolution(AdaptInfo *adaptInfo);  

  virtual void initTimestep(AdaptInfo *adaptInfo);
  virtual void closeTimestep(AdaptInfo *adaptInfo);

  // === getting/setting methods ===

  DOFVector<double> *getPhase()
  {
    return phase;
  }
  
  double getEpsilon()
  {
    return epsilon;
  }
  
  void setPhase(DOFVector<double>* phase_)
  {
    phase = phase_;
  }

  void setDirichletPhase(DOFVector<double>* phase_)
  {
    dirichletPhase = phase_;
  }
  
  void setBcDOF(DOFVector<double>* bcDOF_i, int i)
  {
    bcDOF[i] = bcDOF_i;
  }
  
  void setBcFct(AbstractFunction<WorldVector<double>, WorldVector<double> >* fct_)
  {
    bcFct = fct_;
  }

  void setEpsilon(double epsilon_)
  {
    epsilon = epsilon_;
  }
  
// protected: // methods

  virtual void fillOperators();
  virtual void fillBoundaryConditions();

protected: // variables

  double mu;
  double lambda;
  double rho;
  double epsilon;
  double beta;
  double alpha;
  double dbc_factor;

  WorldVector<double> force;
  DOFVector<WorldVector<double> >* displacement;

  DOFVector<double> *phase;
  DOFVector<double> *phaseOld;

  WorldVector<DOFVector<double>*> bcDOF;
  AbstractFunction<WorldVector<double>, WorldVector<double> >* bcFct;

  DOFVector<double>* dirichletPhase;
  
  FileVectorWriter *fileWriter;
};

struct NeumannBCDof : BinaryAbstractFunction<double, double, WorldVector<double> >
{
  NeumannBCDof() : BinaryAbstractFunction<double, double, WorldVector<double> >(5) {}
  
  double operator()(const double& g, const WorldVector<double>& grdPhase) const
  {
    return g/* * AMDiS::Norm2<double, WorldVector<double> >()(grdPhase)*/;
  }
};

struct NeumannBCFct : BinaryAbstractFunction<double, WorldVector<double>, WorldVector<double> >
{
  NeumannBCFct(AbstractFunction<double, WorldVector<double> >* g_)
  : BinaryAbstractFunction<double, WorldVector<double>, WorldVector<double> >(g_->getDegree() + 3), 
    g(g_) {}
  
  double operator()(const WorldVector<double>& grdPhase, const WorldVector<double>& x) const
  {
    return (*g)(x)/* * AMDiS::Norm2<double, WorldVector<double> >()(grdPhase)*/;
  }
  
private:
  AbstractFunction<double, WorldVector<double> >* g;
};

struct ComponentExtractor_ : AbstractFunction<double, WorldVector<double> >
{
  ComponentExtractor_(AbstractFunction<WorldVector<double>, WorldVector<double> >* fct_, int comp_, double factor_=1.0)
  : AbstractFunction<double, WorldVector<double> >(fct_->getDegree()),
    fct(fct_), 
    comp(comp_),
    factor(factor_) {}
  
  double operator()(const WorldVector<double>& x) const {
    WorldVector<double> result = (*fct)(x);
    return factor*result[comp];
  }
  
private:
  AbstractFunction<WorldVector<double>, WorldVector<double> >* fct;
  int comp;
  double factor;
};

#endif // LINEAR_ELASTICITY_H
