#include "NavierStokes_Chorin.h"

#define UPDATE_VELOCITY_SYSTEM

using namespace std;
using namespace AMDiS;

NavierStokes_Chorin::NavierStokes_Chorin(const std::string &name_, WorldVector<AbstractFunction<double, WorldVector<double> >* > initialVelocityFcts_) :
  ProblemInstatBase(name_,NULL),
  initialVelocityFcts(initialVelocityFcts_),
  predictorProb(NULL), 
  pressureProb(NULL), 
  correctorProb(NULL),
  calcPressure(false),
  forceDBC(false),
  initialVelocityIsSet(false),
  oldMeshChangeIdx(0),
  nTimesteps(-1),
  simpleAlg(2),
  poissonPertubation(0),
  laplaceType(0),
  nonLinTerm(2),
  oldTimestep(0.0),
  pressureStep(0.0),
  viscosity(1.0),
  c(0.0)
{
  dim = Global::getGeo(WORLD);

  // create basic problems
  predictorProb = new ProblemStatType(name + "->predictor");
  pressureProb = new ProblemStatType(name + "->pressure");
  correctorProb = new ProblemStatType(name + "->corrector");
  
  // set Pointer to main problem
  nsProb = predictorProb;

  // type of SIMPLE-algorithm
  Initfile::get(name + "->simple algorithm",simpleAlg);
//   if (simpleAlg < 1 || simpleAlg > 2)
//     throw(std::runtime_error("SIMPLE-algorithm: Possible values in {1,2}!"));

  // modification of the neumann-poisson-problem by additional pertubation or
  // fixing of one dirichlet-node. Can be set to 0 if other dirichlet BC are set
  Initfile::get(name + "->poisson problem pertubation", poissonPertubation);

  // force the homogeniouse dirichlet BC if combination of dirichlet and neumann BC
  // are set and AMDiS can not handle this combination automatically
  Initfile::get(name + "->force dirichlet bc", forceDBC);


  // parameters for navier-stokes
  Initfile::get(name + "->viscosity", viscosity);
  // type of laplace operator: 0... div(nu*grad(u)), 1... div(0.5*nu*(grad(u)+grad(u)^T))
  Initfile::get(name + "->laplace operator", laplaceType); 
  // type of non-linear term: 0... u^old*grad(u_i^old), 1... u'*grad(u_i^old), 2... u^old*grad(u'_i)
  Initfile::get(name + "->non-linear term", nonLinTerm); 

  force.set(0.0);
  Initfile::get(name + "->force", force, DEBUG_INFO);
};


NavierStokes_Chorin::~NavierStokes_Chorin() 
{ FUNCNAME("NavierStokes_Chorin::~NavierStokes_Chorin()");

  if (calcPressure)
    delete pressure;
  if (initialVelocityIsSet) {
    delete initialX;
    delete initialY;
  }
};


void NavierStokes_Chorin::initialize( Flag initFlag,
                                      ProblemStatType *adoptProblem,
                                      Flag adoptFlag)
{ FUNCNAME("NavierStokes_Chorin::initialize()");

  Flag localInitFlag = INIT_ALL - INIT_MESH - CREATE_MESH - INIT_FE_SPACE;
  Flag localAdoptFlag = INIT_MESH + INIT_FE_SPACE;

  predictorProb->initialize(initFlag, adoptProblem, adoptFlag);
  pressureProb->initialize(localInitFlag, predictorProb, localAdoptFlag);
  correctorProb->initialize(localInitFlag, predictorProb, localAdoptFlag);
};


void NavierStokes_Chorin::initTimeInterface()
{
MSG("NavierStokes_Chorin::initTimeInterface()\n");

  for (int i = 0; i < boundaryDOF.getSize(); ++i)
    boundaryDOF[i] = new DOFVector<double>(predictorProb->getFeSpace(i), "boundary_i");

  fillOperators();
  fillBoundaryConditions();
};


void NavierStokes_Chorin::solveInitialProblem(AdaptInfo *adaptInfo) 
{ FUNCNAME("NavierStokes_Chorin::solveInitialProblem()");
MSG("NavierStokes_Chorin::solveInitialProblem()\n");

  if (initialVelocityFcts.size() > 0) {
    initialVelocityIsSet = true;
  
    MSG("interpol initial velocity fcts (%d)...\n",initialVelocityFcts.size());
    for (unsigned i = 0; i < dim; ++i) {
      predictorProb->getSolution()->getDOFVector(i)->interpol(initialVelocityFcts[i]);
      predictorProb->setExactSolutionFct(initialVelocityFcts[i], i);
    }
  }

  // initial pressure
  MSG("initial pressure...\n");
  pressureProb->getSolution()->getDOFVector(0)->set(0.0);
  pressureProb->setExactSolution(pressureProb->getSolution()->getDOFVector(0), 0);

  // initial parameters for detecting mesh changes
  MSG("oldMeshChangeIdx...\n");
  oldMeshChangeIdx= predictorProb->getMesh()->getChangeIndex();

  // calc pressure by normalization and correction of the value
  MSG("pressureDOF...\n");
  Initfile::get(name + "->calculate pressure", calcPressure);
  if (calcPressure)
    pressure= new DOFVector<double>(pressureProb->getFeSpace(0), "pressure");
  
  MSG("[finished]\n");
};


void NavierStokes_Chorin::transferInitialSolution(AdaptInfo *adaptInfo)
{ FUNCNAME("NavierStokes_Chorin::transferInitialSolution()");
MSG("NavierStokes_Chorin::transferInitialSolution()\n");

  for (unsigned i = 0; i < boundaryDOF.size(); ++i)
    boundaryDOF[i]->copy(*predictorProb->getSolution()->getDOFVector(i));

  if (abs(c) > DBL_TOL) {
    INFO(DEBUG_INFO,2)("correct pressure...\n");
    solvePressureProb(adaptInfo);
  
    INFO(DEBUG_INFO,2)("correct velocity...\n");
    solveCorrectorProb(adaptInfo);
  } else
    correctorProb->getSolution()->copy(*predictorProb->getSolution());

  predictorProb->setExactSolution(correctorProb->getSolution()->getDOFVector(0), 0);
  predictorProb->setExactSolution(correctorProb->getSolution()->getDOFVector(1), 1);

//   for (unsigned i = 0; i < boundaryDOF.size(); ++i)
//     boundaryDOF[i]->copy(*correctorProb->getSolution()->getDOFVector(i));

  writeFiles(adaptInfo, false);
};


void NavierStokes_Chorin::beginIteration(AdaptInfo *adaptInfo) 
{ FUNCNAME("NavierStokes_Chorin::beginIteration()");
MSG("NavierStokes_Chorin::beginIteration()\n");
	
  MSG("\n");
  MSG("[[ Navier-Stokes iteration ]]\n");
    
  // assemble some blocks only once, if some conditions are fullfilled
  #ifdef HAVE_PARALLEL_DOMAIN_AMDIS
  bool fixedMatrixCell= false;
  #else
  bool fixedMatrixCell= adaptInfo->getTimestepNumber() > 0
    && std::abs(adaptInfo->getTimestep() - oldTimestep) < DBL_TOL
    && oldMeshChangeIdx == getMesh()->getChangeIndex();
  #endif
  for(unsigned i= 0; i < fixedMatrixTimestep.size(); ++i) {
    predictorProb->setAssembleMatrixOnlyOnce(
      fixedMatrixTimestep[i].first, 
      fixedMatrixTimestep[i].second,
      fixedMatrixCell);
  }
};


Flag NavierStokes_Chorin::solvePredictorProb( AdaptInfo *adaptInfo, 
                                              Flag toDo, 
                                              bool fixedMatrix)
{ 
  FUNCNAME("NavierStokes_Chorin::solvePredictorProb()");
  
  MSG("NavierStokes_Chorin::solvePredictorProb()\n");

  Flag markFlag;

  if (toDo.isSet(BUILD))
    predictorProb->buildAfterCoarsen(adaptInfo, markFlag, !fixedMatrix, true);

  if (toDo.isSet(SOLVE))
    predictorProb->solve(adaptInfo);

  return markFlag;
};


Flag NavierStokes_Chorin::solvePressureProb(AdaptInfo *adaptInfo, 
                                            Flag toDo, 
                                            bool fixedMatrix)
{ 
  FUNCNAME("NavierStokes_Chorin::solvePressureProb()");

  Flag markFlag;

  if (toDo.isSet(BUILD)) {
    pressureProb->buildAfterCoarsen(adaptInfo, markFlag,!fixedMatrix,true);

    SolverMatrix<Matrix<DOFMatrix*> > solverMatrix;
    if (!fixedMatrix && poissonPertubation > 0) {
      INFO(DEBUG_INFO,2)("apply Matrix Pertubation(%d)...\n", poissonPertubation);
      if (poissonPertubation == 1)
        applySingularPertubation(pressureProb->getSystemMatrix(0,0)->getBaseMatrix());
      else if (poissonPertubation == 2) {
        WorldVector<double> p;
        p[0]= 0.0;
        p[1]= -1.0; //dimension[1];
        getMesh()->dofCompress();
        DegreeOfFreedom idx;
	bool inside = pressureProb->getSolution()->getDOFVector(0)->getDofIdxAtPoint(p, idx);

        if (inside) {
          if (pressureProb->getSystemMatrix(0,0) != NULL)
            applyDbc(pressureProb->getSystemMatrix(0,0)->getBaseMatrix(), *pressureProb->getRhsVector(0), idx, true, 0.0);
        }
      }

      solverMatrix.setMatrix(*pressureProb->getSystemMatrix());
      pressureProb->setSolverMatrix(solverMatrix);
    }
  }

  if (toDo.isSet(SOLVE)) {
    pressureProb->solve(adaptInfo);

    if (calcPressure == 1) {
      normalizePressure(pressureProb->getSolution()->getDOFVector(0), 0.0);
      transformDOF(pressure,
        pressureProb->getSolution()->getDOFVector(0),
        pressure, new Add<double>);
    }
  }

  return markFlag;
};


Flag NavierStokes_Chorin::solveCorrectorProb( AdaptInfo *adaptInfo, 
                                              Flag toDo, 
                                              bool fixedMatrix)
{ FUNCNAME("NavierStokes_Chorin::solveCorrectorProb()");

  Flag markFlag;

  #if defined UPDATE_VELOCITY_DIRECT
  DOFVector<double> *pressure_temp;
  if (pressureProb->getFeSpace(0) != correctorProb->getFeSpace(0)) {
    INFO(DEBUG_INFO,2)("create new DOFVector for pressure interpolation...\n");
    pressure_temp = new DOFVector<double>(correctorProb->getFeSpace(0), "pressure");
    pressure_temp->interpol(pressureProb->getSolution()->getDOFVector(0));
  } else {
    INFO(DEBUG_INFO,2)("use exististing DOFVector for pressure...\n");
    pressure_temp = pressureProb->getSolution()->getDOFVector(0);
  }

  WorldVector<DOFVector<double>* > gradPressure;
  for (int i = 0; i < gradPressure.getSize(); ++i)
    gradPressure[i] = new DOFVector<double>(correctorProb->getFeSpace(0), "gradPressure_i");
  DOFVector<WorldVector<double> > gradPressure_temp(correctorProb->getFeSpace(0), "gradPressure");

  pressure_temp->getRecoveryGradient(&gradPressure_temp);
  //     pressure_temp->getGradient(&gradPressure_temp);
  INFO(DEBUG_INFO,2)("transform gradient of DOFVector...\n");
  transform(&gradPressure_temp, &gradPressure);

  for (int i = 0; i < gradPressure.getSize(); ++i) {
    transformDOF(predictorProb->getSolution()->getDOFVector(i),
      gradPressure[i],
      correctorProb->getSolution()->getDOFVector(i), new Subtract<double>);
  }
  
  // delete garbage
  if(pressureProb->getFeSpace(0)!=correctorProb->getFeSpace(0))
    delete pressure_temp;
  for(int i= 0; i<gradPressure.getSize(); ++i)
    delete gradPressure[i];
  #else
  correctorProb->buildAfterCoarsen(adaptInfo, markFlag,!fixedMatrix,true);
  correctorProb->solve(adaptInfo);
  #endif

  return markFlag;
}


Flag NavierStokes_Chorin::oneIteration( AdaptInfo *adaptInfo, 
                                        Flag toDo) 
{ 
  FUNCNAME("NavierStokes_Chorin::oneIteration()");

  Flag flag, markFlag;

  for (int i = 0; i < simpleAlg; i++) {
    MSG("solve predictorProb...\n");
    pressureStep = std::min(1.0, static_cast<double>(i));
    bool fixedMatrixPredictor = adaptInfo->getTimestepNumber() > 0
      && std::abs(adaptInfo->getTimestep() - oldTimestep) < DBL_TOL
      && oldMeshChangeIdx == getMesh()->getChangeIndex();
    solvePredictorProb(adaptInfo, toDo, false);

    MSG("solve pressureProb...\n");
    bool fixedMatrixPressure = oldMeshChangeIdx == getMesh()->getChangeIndex();
    solvePressureProb(adaptInfo, toDo, false);
  }

  MSG("solve correctorProb...\n");
  bool fixedMatrixCorrector = oldMeshChangeIdx == getMesh()->getChangeIndex();
  solveCorrectorProb(adaptInfo, toDo, false);

  if (calcPressure == 1)
    pressureProb->getSolution()->getDOFVector(0)->copy(*pressure);

  oldTimestep= adaptInfo->getTimestep();
  oldMeshChangeIdx= getMesh()->getChangeIndex();

  return flag;
};


void NavierStokes_Chorin::endIteration(AdaptInfo *adaptInfo) 
{ FUNCNAME("NavierStokes_Chorin::endIteration()");
	
  MSG("\n");
  MSG("[[ end of Navier-Stokes iteration ]]\n");
};


void NavierStokes_Chorin::clean()
{ FUNCNAME("NavierStokes_Chorin::clean()");

  // delete problems that are not necessary any more
//   Helpers::cleanProblem(predictorProb);
//   Helpers::cleanProblem(pressureProb);
//   Helpers::cleanProblem(correctorProb, DELETE_ALL - DELETE_SOLUTION);
};


void NavierStokes_Chorin::normalizePressure(DOFVector<double> *p, 
                                            double mean)
{ FUNCNAME("NavierStokes_Chorin::normalizePressure()");

  if (calcPressure != 1)
    throw(std::runtime_error("Flag 'calc pressure' must be set to 1!"));

  double intP = -1.0;
  double area = 1.0;

  if (!Helpers::isNumber(area) || area <= 0.0)
    throw(std::runtime_error("Area must be > 0!\n"));

  int i;
  for (i= 0; abs(intP-mean) > 1.e-8 && i < 10; ++i) {
    intP= p->Int();
    if (!Helpers::isNumber(intP))
      throw(std::runtime_error("pressureProb containes non-valid values!"));
    transformDOF(p,(mean-intP)/area,p, new Add<double>);
  }
  MSG("normalizePressure... %d steps\n",i+1);
};


void NavierStokes_Chorin::fillOperators()
{ FUNCNAME("NavierStokes_Chorin::fillOperators()");
  MSG("NavierStokes_Chorin::fillOperators()\n");

  fillPredictorProb();
  fillPressureProb();
  fillCorrectorProb();
};


void NavierStokes_Chorin::fillPredictorProb()
{ FUNCNAME("NavierStokes_Chorin::fillPredictorProb()");
  MSG("NavierStokes_Chorin::fillPredictorProb()\n");

  WorldVector<DOFVector<double>* > vel;
  for (unsigned i = 0; i < dim; i++)
    vel[i]= correctorProb->getSolution()->getDOFVector(i);

  // fill operators for predictorProb
  for (unsigned i = 0; i < dim; ++i) {
    /// < (1/tau)*u'_i , psi >
    Operator *opTime = new Operator(predictorProb->getFeSpace(i), predictorProb->getFeSpace(i));
    opTime->addZeroOrderTerm(new Simple_ZOT);
    predictorProb->addMatrixOperator(*opTime, i, i, getInvTau());
    /// < (1/tau)*u_i^old , psi >
    opTime->setUhOld(correctorProb->getSolution()->getDOFVector(i));
    predictorProb->addVectorOperator(*opTime, i, getInvTau());
 
    if (nonLinTerm == 0) {
      /// < u^old*grad(u_i^old) , psi >
      Operator *opUGradU = new Operator(predictorProb->getFeSpace(i),predictorProb->getFeSpace(i));
      opUGradU->addFirstOrderTerm(new WorldVec_FOT(vel, -1.0), GRD_PHI);
      opUGradU->setUhOld(correctorProb->getSolution()->getDOFVector(i));
      predictorProb->addVectorOperator(*opUGradU, i);
    } else if (nonLinTerm == 1) {
      /// < u'*grad(u_i^old) , psi >
      for (unsigned j = 0; j < dim; ++j) {
        Operator *opUGradU = new Operator(predictorProb->getFeSpace(i),predictorProb->getFeSpace(i));
        opUGradU->addZeroOrderTerm(new PartialDerivative_ZOT(
          correctorProb->getSolution()->getDOFVector(i), j));
        predictorProb->addMatrixOperator(*opUGradU, i, j);
      }
    } else {
      /// < u^old*grad(u'_i) , psi >
      for(unsigned j = 0; j < dim; ++j) {
        Operator *opUGradU = new Operator(predictorProb->getFeSpace(i),predictorProb->getFeSpace(i));
        opUGradU->addFirstOrderTerm(new VecAndPartialDerivative_FOT(
          correctorProb->getSolution()->getDOFVector(j), j), GRD_PHI);
        predictorProb->addMatrixOperator(*opUGradU, i, i);
      }           
    }

    /// Diffusion-Operator (including Stress-Tensor for space-dependent viscosity
    addLaplaceTerm(i);
  
    /// < p , d_i(psi) > TODO: Add boundary-integral for neumann-boundary
    Operator *opGradP2 = new Operator(predictorProb->getFeSpace(i),pressureProb->getFeSpace(0));
    opGradP2->addFirstOrderTerm(new PartialDerivative_FOT(i, -1.0), GRD_PHI);
    opGradP2->setUhOld(pressureProb->getSolution()->getDOFVector(0));
    predictorProb->addVectorOperator(*opGradP2, i, getPressureStep());
  
    /// external force, i.e. gravitational force
    if (norm(force) > DBL_TOL) {
      Operator *opForce = new Operator(predictorProb->getFeSpace(i), predictorProb->getFeSpace(i));
      opForce->addZeroOrderTerm(new Simple_ZOT(force[i]));
      predictorProb->addVectorOperator(*opForce, i);
    }
  }
};


void NavierStokes_Chorin::fillPressureProb()
{ FUNCNAME("NavierStokes_Chorin::fillPressureProb()");
  MSG("NavierStokes_Chorin::fillPressureProb()\n");

  // preassure-calculation: div(grad(p)) = 1/tau*div(u')
  // -------------------------------------------------------------
  for(unsigned i = 0; i < dim; ++i) {
    /// < (1/tau)*d_i(u'_i) , psi >
    Operator *opDivU = new Operator(pressureProb->getFeSpace(0),predictorProb->getFeSpace(i));
    opDivU->addZeroOrderTerm(new PartialDerivative_ZOT(
      predictorProb->getSolution()->getDOFVector(i), i));
    pressureProb->addVectorOperator(*opDivU, 0, getInvTau());
  }
  /// < -grad(p) , grad(psi) >
  Operator *laplaceP = new Operator(pressureProb->getFeSpace(0), pressureProb->getFeSpace(0));
  laplaceP->addSecondOrderTerm(new Simple_SOT(-1.0));
  pressureProb->addMatrixOperator(*laplaceP, 0, 0);
};


void NavierStokes_Chorin::fillCorrectorProb()
{ FUNCNAME("NavierStokes_Chorin::fillCorrectorProb()");
  MSG("NavierStokes_Chorin::fillCorrectorProb()\n");

  // corrector step: u = u' - tau*grad(p)
  // ---------------------------------------
  for(unsigned i = 0; i < dim; ++i) {
    /// < u_i , psi >
    Operator *opSimple = new Operator(correctorProb->getFeSpace(i),correctorProb->getFeSpace(i));
    opSimple->addZeroOrderTerm(new Simple_ZOT);
    correctorProb->addMatrixOperator(*opSimple, i, i);
    /// < u'_i , psi >
    opSimple->setUhOld(predictorProb->getSolution()->getDOFVector(i));
    correctorProb->addVectorOperator(*opSimple, i);
    /// < -tau*d_i(p) , psi >
    Operator *opGradQ = new Operator(correctorProb->getFeSpace(i),pressureProb->getFeSpace(0));
    opGradQ->addFirstOrderTerm(new PartialDerivative_FOT(i, -1.0), GRD_PHI);
    opGradQ->setUhOld(pressureProb->getSolution()->getDOFVector(0));
    correctorProb->addVectorOperator(*opGradQ, i, getTau());
  }
};


void NavierStokes_Chorin::addLaplaceTerm(int i)
{ FUNCNAME("NavierStokes_Chorin::addLaplaceTerm()");
  MSG("NavierStokes_Chorin::addLaplaceTerm(%d)\n",i);

    /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
    if (laplaceType == 1) {
      for (unsigned j = 0; j < dim; ++j) {
        Operator *opLaplaceUi1 = new Operator(predictorProb->getFeSpace(i), predictorProb->getFeSpace(j));
        opLaplaceUi1->addSecondOrderTerm(new FactorIJ_SOT(j, i, viscosity));
        predictorProb->addMatrixOperator(*opLaplaceUi1, i, j);
      }
    }
    
    /// < alpha*grad(u'_i) , grad(psi) >
    Operator *opLaplaceUi = new Operator(predictorProb->getFeSpace(i), predictorProb->getFeSpace(i));
      opLaplaceUi->addSecondOrderTerm(new Simple_SOT(viscosity));
    predictorProb->addMatrixOperator(*opLaplaceUi, i, i);
};


/// TODO: Boundary-Conditions verallgemeinern
void NavierStokes_Chorin::fillBoundaryConditions()
{ FUNCNAME("NavierStokes_Chorin::fillBoundaryConditions()");
  MSG("NavierStokes_Chorin::fillBoundaryConditions()\n");

  DOFVector<double> *zeroDOF = new DOFVector<double>(predictorProb->getFeSpace(0), "zero");
  zeroDOF->set(0.0);

  /// at rigid wall: no-slip boundary condition
  for (unsigned j = 0; j < dim; ++j) {
    predictorProb->addDirichletBC(LINKS, j, j, boundaryDOF[j]);
    predictorProb->addDirichletBC(RECHTS, j, j, boundaryDOF[j]);
  }
  
  /// at inflow: prescribed velocity
  for (unsigned j = 0; j < dim; ++j) {
   predictorProb->addDirichletBC(OBEN, j, j, boundaryDOF[j]);
//    predictorProb->addDirichletBC(OBEN, j, j, predictorProb->getExactSolution(j)); /// test: Dirichlet-BC at outflow boundary
  }
  
  /// at outflow: constant pressure
  pressureProb->addDirichletBC(UNTEN, 0,0, zeroDOF);
//   pressureProb->addDirichletBC(OBEN, 0,0, zeroDOF);


  /// boundaryManager wird von Diagonale kopiert, wenn MatrixOperator hinzugefuegt wird -> erst MatrixOperatoren, danach erst BoundaryOperatoren
//   for (unsigned i = 0; i < dim; ++i) {
//     for (unsigned j = 0; j < dim; ++j) {
//       Operator *opLaplaceBoundary = new Operator(predictorProb->getFeSpace(i), predictorProb->getFeSpace(j));
//       opLaplaceBoundary->addFirstOrderTerm(new ViscosityPartialDerivativeBoundary_FOT(
//         chProb.getSolution()->getDOFVector(0), i, j, viscosity1, viscosity2), GRD_PHI);
//       for (unsigned b= 0; b < nsNeumannBC.size(); ++b) {
//         if (nsNeumannBC[b] != 0) {
//           predictorProb->addBoundaryMatrixOperator(nsNeumannBC[b], opLaplaceBoundary, i, j);
//         }
//       }
//     }
//   }
};


void NavierStokes_Chorin::initTimestep(AdaptInfo *adaptInfo)
{ FUNCNAME("NavierStokes_Chorin::initTimestep()");

  WorldVector<DOFVector<double>* > gradPressure;
  for (int i = 0; i < gradPressure.getSize(); ++i)
    gradPressure[i] = new DOFVector<double>(correctorProb->getFeSpace(0), "gradPressure_i");
  DOFVector<WorldVector<double> > gradPressure_temp(correctorProb->getFeSpace(0), "gradPressure");

  pressureProb->getSolution()->getDOFVector(0)->getRecoveryGradient(&gradPressure_temp);
  transform(&gradPressure_temp, &gradPressure);

  for (int i = 0; i < gradPressure.getSize(); ++i) {
    transformDOF(predictorProb->getExactSolution(i),
      gradPressure[i],
      boundaryDOF[i], new AddFactor<double>(*getTau()));
  }
  
  // delete garbage
  for(int i= 0; i<gradPressure.getSize(); ++i)
    delete gradPressure[i];
};


void NavierStokes_Chorin::closeTimestep(AdaptInfo *adaptInfo)
{ FUNCNAME("NavierStokes_Chorin::closeTimestep()");

  int outputPeriod = 1;
  Initfile::get("user parameter->write every ith timestep", outputPeriod, DEBUG_INFO);
  if (adaptInfo->getTimestepNumber() % outputPeriod == 0
    && adaptInfo->getStartTime() < adaptInfo->getEndTime()) {
    writeFiles(adaptInfo, false);
  }
};


void NavierStokes_Chorin::writeFiles(AdaptInfo *adaptInfo, bool force)
{
  predictorProb->writeFiles(adaptInfo, force);
  pressureProb->writeFiles(adaptInfo, force);
  correctorProb->writeFiles(adaptInfo, force);
};
