#include "AMDiS.h"
#include "movingMesh.h"

#include "boost/date_time/posix_time/posix_time.hpp"

using namespace AMDiS;
using namespace boost::posix_time;

///_______________________________________________________________________________________________________________________

class Elasticity_Fluid : public StandardBaseProblem
{
public:
  typedef StandardBaseProblem super;

public:
  Elasticity_Fluid(std::string name_) : super(name_), displacement(NULL) {}

  ~Elasticity_Fluid()
  { }
  
  void setDisplacement(AbstractFunction<double, WorldVector<double> >* displacement_)
  {
	displacement = displacement_;
  }

  void initData()
  {
    super::initData();
    dim = getMesh()->getDim();
	
	displacementDOF = new DOFVector<double>(getFeSpace(), "displacement_fluid");
  }
  
  void fillOperators()
  {
    // ===== create matrix operator =====
    for (size_t i = 0; i < dow; i++) {
	  Operator *matrixOperator = new Operator(getFeSpace());
	  matrixOperator->addTerm(new Simple_SOT);
	  prob->addMatrixOperator(matrixOperator, i, i);
	  
// 	  Operator *rhsOperator = new Operator(getFeSpace());
// 	  rhsOperator->addTerm(new CoordsAtQP_ZOT(new AMDiS::Const<double, WorldVector<double> >(1.0)));
// 	  prob->addVectorOperator(rhsOperator, i);
	}
  } 

  void fillBoundaryConditions()
  {
	MSG("Elasticity_Fluid::fillBoundaryConditions\n");
    super::fillBoundaryConditions();

    AbstractFunction<double, WorldVector<double> > *zero = new AMDiS::Const<double, WorldVector<double> >(0.0);
    double zeroValue = 0.0;

    //  +--+
    //  |  |   1
    //10|  +==========
    //  |  |
    //  +--+

    for (int i = 0; i < dow; i++) {
      prob->addDirichletBC(10, i, i, zero);
      prob->addDirichletBC(2, i, i, zero);
      prob->addDirichletBC(3, i, i, zero);
      prob->addDirichletBC(4, i, i, zero);
      prob->addDirichletBC(5, i, i, zero);
    }
	prob->addDirichletBC(1, 0, 0, zero);
	prob->addDirichletBC(1, 1, 1, displacementDOF);
  }
  
  void initTimestep(AdaptInfo *adaptInfo)
  { MSG("initTimestep()\n");
  
    super::initTimestep(adaptInfo);
	displacementDOF->interpol(displacement);
  }
  
  void closeTimestep(AdaptInfo *adaptInfo)
  { MSG("closeTimestep()\n");
  
    super::closeTimestep(adaptInfo);
	VtkWriter::writeFile(displacementDOF, "displacement_fluid.vtu");
  }

private:
  AbstractFunction<double, WorldVector<double> >* displacement;
  DOFVector<double>* displacementDOF;
};


///_______________________________________________________________________________________________________________________

class Elasticity_Solid : public StandardBaseProblem
{
public:
  typedef StandardBaseProblem super;

public:
  Elasticity_Solid(std::string name_) : super(name_), displacement(NULL) {}

  ~Elasticity_Solid()
  { }

  void initData()
  {
    super::initData();
    dim = getMesh()->getDim();
	
	displacementDOF = new DOFVector<double>(getFeSpace(), "displacement_fluid");
  }
  
  void setDisplacement(AbstractFunction<double, WorldVector<double> >* displacement_)
  {
	displacement = displacement_;
  }
  
  void fillOperators()
  {
    // ===== create matrix operator =====
    for (size_t i = 0; i < dow; i++) {
	  Operator *matrixOperator = new Operator(getFeSpace());
	  matrixOperator->addTerm(new Simple_SOT);
	  prob->addMatrixOperator(matrixOperator, i, i);
	  
// 	  Operator *rhsOperator = new Operator(getFeSpace());
// 	  rhsOperator->addTerm(new CoordsAtQP_ZOT(new AMDiS::Const<double, WorldVector<double> >(1.0)));
// 	  prob->addVectorOperator(rhsOperator, i);
	}
  }

  void fillBoundaryConditions()
  {
	MSG("Elasticity_Solid::fillBoundaryConditions\n");
    super::fillBoundaryConditions();

    AbstractFunction<double, WorldVector<double> > *zero = new AMDiS::Const<double, WorldVector<double> >(0.0);
    double zeroValue = 0.0;

    //  +--+
    //  |  |   1
    //10|  +==========
    //  |  |
    //  +--+

    for (int i = 0; i < dow; i++) {
      prob->addDirichletBC(10, i, i, zero);
      prob->addDirichletBC(2, i, i, zero);
      prob->addDirichletBC(3, i, i, zero);
      prob->addDirichletBC(4, i, i, zero);
      prob->addDirichletBC(5, i, i, zero);
    }
	prob->addDirichletBC(1, 0, 0, zero);
	prob->addDirichletBC(1, 1, 1, displacementDOF);
  }
  
  void initTimestep(AdaptInfo *adaptInfo)
  { MSG("initTimestep()\n");
  
    super::initTimestep(adaptInfo);
	displacementDOF->interpol(displacement);
  }
  
  void closeTimestep(AdaptInfo *adaptInfo)
  { MSG("closeTimestep()\n");
  
    super::closeTimestep(adaptInfo);
	VtkWriter::writeFile(displacementDOF, "displacement_solid.vtu");
  }

private:
  AbstractFunction<double, WorldVector<double> >* displacement;
  DOFVector<double>* displacementDOF;
};


int main(int argc, char** argv)
{ FUNCNAME("main");

  AMDiS::init(argc, argv);

  Elasticity_Fluid elastProb("elasticity->fluid");
  Elasticity_Solid elastProb2("elasticity->solid");
  
  MovingMesh<Elasticity_Fluid, Elasticity_Solid> mainProb("main", &elastProb, &elastProb2);

  // Adapt-Infos
  AdaptInfo adaptInfo("adapt", mainProb.getNumComponents());
  mainProb.initialize(&adaptInfo);

  // adaption loop - solve ch-prob and ns-prob
  AdaptInstationary adaptInstat("adapt", mainProb, adaptInfo, mainProb, adaptInfo);
  
  ptime start_time = microsec_clock::local_time();
  int error_code = adaptInstat.adapt(); 
  time_duration td = microsec_clock::local_time()-start_time;

  MSG("elapsed time= %d sec\n", td.total_seconds());

  AMDiS::finalize();

  return error_code;
};
