//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.


#include "time/ExtendedRosenbrockStationary.h"
#include "io/VtkWriter.h"
#include "ExtendedRosenbrockStationary.h"
#include "SystemVector.h"
#include "OEMSolver.h"
#include "Debug.h"
#include "POperators_ZOT.h"

#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
#include "parallel/MeshDistributor.h"
#endif

using namespace AMDiS;

  void ExtendedRosenbrockStationary::acceptTimestep()
  {
    *solution = *newUn;
    *unVec = *newUn;
  }


  void ExtendedRosenbrockStationary::init()
  {
    stageSolution = new SystemVector(*solution);
    unVec = new SystemVector(*solution);
    timeRhsVec = new SystemVector(*solution);
    newUn = new SystemVector(*solution);
    tmp = new SystemVector(*solution);
    if (!fixedTimestep)
      lowSol = new SystemVector(*solution);    

    stageSolution->set(0.0);
    unVec->set(0.0);
    
    stageSolutions.resize(rm->getStages());
    for (int i = 0; i < rm->getStages(); i++) {
      stageSolutions[i] = new SystemVector(*solution);
      stageSolutions[i]->set(0.0);
    }
  }


  void ExtendedRosenbrockStationary::buildAfterCoarsen(AdaptInfo *adaptInfo, Flag flag,
					       bool asmMatrix, bool asmVector)
  {        
    FUNCNAME("ExtendedRosenbrockStationary::buildAfterCoarsen()");

    TEST_EXIT(tauPtr)("No tau pointer defined in stationary problem!\n");

    if (first) {
      first = false;
      *unVec = *solution;
    }
    
    *newUn = *unVec;    
    if (!fixedTimestep)
      *lowSol = *unVec;

    for (int i = 0; i < rm->getStages(); i++) {
      
      stageTime = (*oldTime) + rm->getAlphaI(i) * (*tauPtr);
      tauGammaI = rm->getGammaI(i) * (*tauPtr);
      
      // stage-solution: u_s(i) = u_old + sum_{j=0}^{i-1} a_ij*U_j
      *stageSolution = *unVec;
      for (int j = 0; j < i; j++) {
	*tmp = *(stageSolutions[j]);
	*tmp *= rm->getA(i, j);
	*stageSolution += *tmp;
      }

      // Dirichlet-BC implemented as additional algebraic equation u = g(x,t) on boundary
      // => U_i = -u_s(i) + g(x,t_s(i)) + tau*gamma_i* d_t(g)(t_old) on boundary
      // where u_s(i) = ith stage-solution, t_s(i) = ith stage-time
      for (unsigned int j = 0; j < boundaries.size(); j++) {
	boundaries[j].vec->interpol(boundaries[j].fct);
	*(boundaries[j].vec) -= *(stageSolution->getDOFVector(boundaries[j].col));
	
	if (boundaries[j].fctDt != NULL) {
	// time derivative of dirichlet bc is given
	  DOFVector<double>* tmpDt = new DOFVector<double>(getFeSpace(boundaries[j].col), "tmp");
	  tmpDt->interpol(boundaries[j].fctDt);
	  *tmpDt *= tauGammaI;
	  *(boundaries[j].vec) += *tmpDt; // alternativ transformDOF(...) benutzen, für x = x + a*y 
	  delete tmpDt;
	}
      }

      timeRhsVec->set(0.0);
      for (int j = 0; j < i; j++) {
	*tmp = *(stageSolutions[j]);
	*tmp *= (rm->getC(i, j) / *tauPtr);
	*timeRhsVec += *tmp;
      }

      super::buildAfterCoarsen(adaptInfo, flag | UPDATE_DIRICHLET_BC, (i == 0), true);
      
//       mtl::io::matrix_market_ostream out("matrix.mat");
//       SolverMatrix<Matrix<DOFMatrix*> > solverMatrix;
//       solverMatrix.setMatrix(*getSystemMatrix());
//       out << solverMatrix.getMatrix();
//       out.close();
      
      super::solve(adaptInfo, i == 0, true);

      *(stageSolutions[i]) = *solution;
      
      *tmp = *solution;
      *tmp *= rm->getM1(i);
      *newUn += *tmp;

      if (!fixedTimestep) {
	// lower order approximation, with coefficients of embedded method
	// used for local error estimator
	*tmp = *solution;
	*tmp *= rm->getM2(i);
	*lowSol += *tmp;
      }
    }
    
    stageTime = (*oldTime) + (*tauPtr);
    
    if (!fixedTimestep) {
      for (int i = 0; i < nComponents; i++) {
	(*(lowSol->getDOFVector(i))) -= (*(newUn->getDOFVector(i)));
	adaptInfo->setTimeEstSum(lowSol->getDOFVector(i)->l2norm(), i+componentShift);
      }
    }
  }


  void ExtendedRosenbrockStationary::solve(AdaptInfo *adaptInfo, bool, bool)
  {}


  void ExtendedRosenbrockStationary::addOperator(Operator &op, int row, int col,
					 double *factor, double *estFactor)
  {
    FUNCNAME("ExtendedRosenbrockStationary::addOperator()");

    TEST_EXIT(op.getUhOld() == NULL)("UhOld is not allowed to be set!\n");

    op.setUhOld(stageSolution->getDOFVector(col));
    super::addVectorOperator(op, row, factor, estFactor);
  }
  

  void ExtendedRosenbrockStationary::addJacobianOperator(Operator &op, int row, int col,
						 double *factor, double *estFactor)
  {
    FUNCNAME("ExtendedRosenbrockStationary::addJacobianOperator()");
    
    TEST_EXIT(factor == NULL)("Not yet implemented!\n");
    TEST_EXIT(estFactor == NULL)("Not yet implemented!\n");

    super::addMatrixOperator(op, row, col, &minusOne, &minusOne);
  }


  void ExtendedRosenbrockStationary::addTimeOperator(int row, int col, double factor)
  {
    FUNCNAME("ExtendedRosenbrockStationary::addTimeOperator()");

    TEST_EXIT(invTauGamma)("This should not happen!\n");

    Operator *op = new Operator(componentSpaces[row], componentSpaces[col]);
    op->addTerm(new Simple_ZOT(factor));
    super::addMatrixOperator(op, row, col, invTauGamma, invTauGamma);

    Operator *opRhs = new Operator(componentSpaces[row]);
    opRhs->addTerm(new Phase_ZOT(timeRhsVec->getDOFVector(col), factor));
    super::addVectorOperator(opRhs, row);
  }

  
  void ExtendedRosenbrockStationary::addDirichletBC(BoundaryType type, int row, int col,
					    AbstractFunction<double, WorldVector<double> > *fct,
					    AbstractFunction<double, WorldVector<double> > *fctDt)
  {
    FUNCNAME("ExtendedRosenbrockStationary::addDirichletBC()");

    DOFVector<double>* vec = new DOFVector<double>(componentSpaces[col], "vec");
    MyRosenbrockBoundary bound(fct, fctDt, vec, row, col);
    boundaries.push_back(bound);

    super::addDirichletBC(type, row, col, vec);
  }
  

  void ExtendedRosenbrockStationary::addSingularDirichletBC(WorldVector<double> &pos, int row, int col,
							    AbstractFunction<double, WorldVector<double> > &fct)
  {
    DOFVector<double>* vec = new DOFVector<double>(componentSpaces[col], "vec");
    MyRosenbrockBoundary bound(&fct, NULL, vec, row, col);
    boundaries.push_back(bound);

    super::addSingularDirichletBC(pos, row, col, *vec);
  }


  void ExtendedRosenbrockStationary::addSingularDirichletBC(DegreeOfFreedom idx, int row, int col,
							    AbstractFunction<double, WorldVector<double> > &fct)
  {
    DOFVector<double>* vec = new DOFVector<double>(componentSpaces[col], "vec");
    MyRosenbrockBoundary bound(&fct, NULL, vec, row, col);
    boundaries.push_back(bound);

    super::addSingularDirichletBC(idx, row, col, *vec);
  }
  

  void ExtendedRosenbrockStationary::addImplicitDirichletBC(AbstractFunction<double, WorldVector<double> > &signedDist, int row, int col,
							    AbstractFunction<double, WorldVector<double> > &fct)
  {
    DOFVector<double>* vec = new DOFVector<double>(componentSpaces[col], "vec");
    MyRosenbrockBoundary bound(&fct, NULL, vec, row, col);
    boundaries.push_back(bound);

    super::addImplicitDirichletBC(signedDist, row, col, *vec);
  }


  void ExtendedRosenbrockStationary::addImplicitDirichletBC(DOFVector<double> &signedDist, int row, int col,
							    AbstractFunction<double, WorldVector<double> > &fct)
  {
    DOFVector<double>* vec = new DOFVector<double>(componentSpaces[col], "vec");
    MyRosenbrockBoundary bound(&fct, NULL, vec, row, col);
    boundaries.push_back(bound);

    super::addImplicitDirichletBC(signedDist, row, col, *vec);
  }


  void ExtendedRosenbrockStationary::addManualDirichletBC(BoundaryType nr, AbstractFunction<bool, WorldVector<double> >* meshIndicator,
							  int row, int col,
							  AbstractFunction<double, WorldVector<double> > &fct)
  {
    DOFVector<double>* vec = new DOFVector<double>(componentSpaces[col], "vec");
    MyRosenbrockBoundary bound(&fct, NULL, vec, row, col);
    boundaries.push_back(bound);

    super::addManualDirichletBC(nr, meshIndicator, row, col, *vec);
  }


  void ExtendedRosenbrockStationary::addManualDirichletBC(AbstractFunction<bool, WorldVector<double> >* meshIndicator,
							  int row, int col,
							  AbstractFunction<double, WorldVector<double> > &fct)
  {
    DOFVector<double>* vec = new DOFVector<double>(componentSpaces[col], "vec");
    MyRosenbrockBoundary bound(&fct, NULL, vec, row, col);
    boundaries.push_back(bound);

    super::addManualDirichletBC(meshIndicator, row, col, *vec);
  }
